# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import numpy as np
import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import (
    SingleInputSingleOutputParams,
    StandardScaleSkipZerosParams,
)
from kamae.spark.transformers.standard_scale import StandardScaleParams
from kamae.spark.utils.transform_utils import single_input_single_output_array_transform
from kamae.tensorflow.layers import ConditionalStandardScaleLayer

from .base import BaseTransformer


class ConditionalStandardScaleTransformer(
    BaseTransformer,
    StandardScaleParams,
    StandardScaleSkipZerosParams,
    SingleInputSingleOutputParams,
):
    """
    Conditional standard scaler transformer for use in Spark pipelines.
    This is used to standardize/transform the input column using the mean and
    the standard deviation.
    The skip_zeros parameter allows to apply the standard scaling process
    only when input is not equal to zero. If equal to zero, it will remain zero in
    the output value as it was in the input value.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        layerName: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        mean: Optional[List[float]] = None,
        stddev: Optional[List[float]] = None,
        skipZeros: bool = False,
        epsilon: float = 0,
    ) -> None:
        """
        Initializes a ConditionalStandardScaleParams transformer.
        It differs from the default StandardScaleParams in that it gives
        more control over the standard scaling process by allowing the user
        to specify a mask to be used during the fit and transform process, and
        to specify whether to skip zeros during the transform process.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param mean: List of mean values corresponding to the input column.
        :param stddev: List of standard deviation values corresponding to the
        input column.
        :param skipZeros: If True, during spark transform and keras inference,
        do not apply the scaling when the values to scale are equal to zero.
        :param epsilon: Small value to add to conditional check of zeros. Valid only
        when skipZeros is True. Defaults to 0.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(skipZeros=False, epsilon=0)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset using the mean and standard deviation
        to standardize the input column. If a mask value is set, it is used
        to ignore elements in the dataset with that value, and they will remain
        unchanged in the standardization process. If skipZeros is set to True,
        it also ignores elements with value equal to zero in the standardization
        process.

        :param dataset: Pyspark dataframe to transform.
        :returns: Pyspark dataframe with the input column standardized,
         named as the output column.
        """
        original_input_datatype = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(original_input_datatype, ArrayType):
            input_col = F.array(F.col(self.getInputCol()))
            input_datatype = ArrayType(original_input_datatype)
        else:
            input_col = F.col(self.getInputCol())
            input_datatype = original_input_datatype

        shift = F.array([F.lit(m) for m in self.getMean()])
        scale = F.array([F.lit(1.0 / s if s != 0 else 0.0) for s in self.getStddev()])
        if self.getSkipZeros():
            eps = self.getEpsilon()
            func = lambda x: F.transform(  # noqa: E731
                x,
                lambda y, i: F.when(
                    # x != (0 +- eps)
                    F.abs(y) > F.lit(eps),
                    (y - F.lit(shift)[i]) * F.lit(scale)[i],
                ).otherwise(0),
            )
        else:
            func = lambda x: F.transform(  # noqa: E731
                x,
                lambda y, i: (y - F.lit(shift)[i]) * F.lit(scale)[i],
            )
        output_col = single_input_single_output_array_transform(
            input_col=input_col,
            input_col_datatype=input_datatype,
            func=func,
        )
        if not isinstance(original_input_datatype, ArrayType):
            output_col = output_col.getItem(0)
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the standard scaler transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
         that performs the standardization.
        """
        np_mean = np.array(self.getMean())
        np_variance = np.array(self.getStddev()) ** 2
        return ConditionalStandardScaleLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            mean=np_mean,
            variance=np_variance,
            skip_zeros=self.getSkipZeros(),
            epsilon=self.getEpsilon(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class CurrentUnixTimestampLayer(BaseLayer):
    """
    Returns the current unix timestamp in either seconds or milliseconds.

    NOTE: Parity between this and its Spark counterpart is very difficult at the
    millisecond level. TensorFlow provides much more precision of the timestamp,
    and has floating 64-bit precision of the unix timestamp in seconds.
    Whereas Spark 3.4.0 only supports millisecond precision (3 decimal places of unix
    timestamp in seconds). Therefore, parity is not guaranteed at this precision.

    It is recommended not to rely on parity at the millisecond level.
    """

    def __init__(
        self,
        name=None,
        input_dtype: str = None,
        output_dtype: str = None,
        unit: str = "s",
        **kwargs,
    ) -> None:
        """
        Initialises an instance of the CurrentUnixTimestampLayer layer.

        :param name: Name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if unit not in ["milliseconds", "seconds", "ms", "s"]:
            raise ValueError(
                """Unit must be one of ["milliseconds", "seconds", "ms", "s"]"""
            )
        if unit == "milliseconds":
            unit = "ms"
        elif unit == "seconds":
            unit = "s"
        self.unit = unit

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer. Returns `None` as the layer
        only returns the current date as a string. It does not transform any input.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> tf.Tensor:
        """
        Returns the current unix timestamp in either seconds or milliseconds.
        Uses the input tensor to determine the shape of the output tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to determine the shape of the output tensor.
        :returns: The current timestamp tensor in yyyy-MM-dd format.
        """
        current_timestamp_in_seconds = tf.fill(tf.shape(inputs), tf.timestamp())
        return (
            current_timestamp_in_seconds
            if self.unit == "s"
            else current_timestamp_in_seconds * 1000.0
        )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the CurrentUnixTimestamp layer.
        Used for saving and loading from a model.

        Specifically adds the `unit` parameter to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()

        config.update(
            {
                "unit": self.unit,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Tuple

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql.types import ArrayType, DataType


def get_array_nesting_level(column_dtype: DataType, start_level: int = 0) -> int:
    """
    Calls get_array_nesting_level_and_element_dtype to determine the nesting level of a
    Spark array column.

    :param column_dtype: DataType of the column to check.
    :param start_level: Counter of the number of times it had to recurse to get the
    element type of the array column. Default is 0.
    :returns: The nesting level of the array column.
    """
    return get_array_nesting_level_and_element_dtype(column_dtype, start_level)[0]


def get_element_type(dtype: DataType) -> DataType:
    """
    Gets the element type of given datatype. If the datatype is not an array,
    it returns the datatype itself.

    :param dtype: The datatype to get the element type of.
    :returns: The element type of the datatype.
    """
    return get_array_nesting_level_and_element_dtype(dtype)[1]


def get_array_nesting_level_and_element_dtype(
    column_dtype: DataType, start_level: int = 0
) -> Tuple[int, DataType]:
    """
    Determines the nesting level of a Spark array column. Also returns the underlying
    type of the column. Recursively attempts to get the element type of the array column
    until it is no longer an array. Keeps a counter of the number of times it had to
    recurse to get the element type. This counter is the nesting level of the array
    column. Returns this counter along with the last found element type.

    :param column_dtype: DataType of the column to check.
    :param start_level: Counter of the number of times it had to recurse to get the
    element type of the array column. Default is 0.
    :returns: Tuple of the nesting level of the array column and the element type of the
    array column.
    """
    try:
        child_dtype = column_dtype.elementType
    except AttributeError:
        return start_level, column_dtype
    else:
        return get_array_nesting_level_and_element_dtype(
            column_dtype=child_dtype, start_level=start_level + 1
        )


def broadcast_scalar_column_to_array(
    scalar_column: Column, array_column: Column, array_column_datatype: DataType
) -> Column:
    """
    Broadcasts a scalar column to a (possibly nested) array column. Repeats the scalar
    value to match the shape of the given array column.

    :param scalar_column: Spark column to broadcast.
    :param array_column: Spark array column to broadcast to.
    :param array_column_datatype: DataType of the array column. Used to understand how
    nested the array column is.
    :returns: Spark array column with the scalar column broadcasted to it.
    """
    nested_level = get_array_nesting_level(array_column_datatype)
    nested_transform_func = nested_transform(
        func=lambda x: scalar_column, nest_level=nested_level
    )
    return nested_transform_func(array_column)


def broadcast_scalar_column_to_array_with_inner_singleton_array(
    scalar_column: Column, array_column: Column, array_column_datatype: DataType
) -> Column:
    """
    Similar to broadcast_scalar_column_to_array, but does not repeat the scalar value in
    the final innermost array.

    Broadcasts a scalar column to a (possibly nested) array column. Repeats the scalar
    value to match the shape of the first N-1 of an N-dim array. The final innermost
    array is left as a singleton array. Used for the array concatenate operation where
    we do not want to repeat the value in the final array.

    Example:
    scalar_column = 1
    array_column = [[1, 2], [3, 4]]
    array_column_datatype = ArrayType(ArrayType(IntegerType()))

    Output: [[1], [1]]

    Using `broadcast_scalar_column_to_array` with the same example would return

    Output: [[1, 1], [1, 1]]

    :param scalar_column: Spark column to broadcast.
    :param array_column: Spark array column to broadcast to.
    :param array_column_datatype: DataType of the array column. Used to understand how
    nested the array column is.
    :returns: Spark array column with the scalar column broadcasted to it, and with a
    singleton array in the innermost position.
    """
    nested_level = get_array_nesting_level(array_column_datatype)
    nested_transform_func = nested_transform(
        func=lambda x: F.array(scalar_column), nest_level=nested_level - 1
    )
    return nested_transform_func(array_column)


def nested_arrays_zip(
    columns: List[Column], nest_level: int, column_names: Optional[List[str]] = None
) -> Column:
    """
    Zips multiple columns of (possibly nested) array type.

    :param columns: List of Spark array columns to be zipped.
    :param nest_level: Nesting level of the array columns.
    :param column_names: Optional list of column names to use for the zipped array. If
    provided, the final array of structs will use these as the names of the columns.
    If not set, the columns will be named "input_0", "input_1", etc.
    If set, must be the same length as the number of columns.
    :returns: Zipped array column.
    """
    if column_names is not None:
        if len(column_names) != len(columns):
            raise ValueError(
                f"""column_names must be the same length as the number of columns.
                Received {len(column_names)} column names for {len(columns)} columns."""
            )
    else:
        column_names = [f"input_{i}" for i in range(len(columns))]

    # First collect everything into a struct, so we can operate on a single column
    arrays_zipped = F.struct(*[c.alias(n) for n, c in zip(column_names, columns)])
    # For each successive nesting level below this, zip the arrays together, each time
    # pushing the struct down one level.
    for n_lvl in range(0, nest_level):
        arrays_zipped = nested_transform(
            func=lambda x: F.arrays_zip(*[x[n].alias(n) for n in column_names]),
            nest_level=n_lvl,
        )(arrays_zipped)

    return arrays_zipped


def nested_transform(
    func: Callable[[Column], Column], nest_level: int
) -> Callable[[Column], Column]:
    """
    Creates a nested version of `pyspark.sql.functions.transform` that can be used to
    apply a given `func` elementwise on a nested array column.

    :param func: Function to apply to each element of the nested array column.
    :param nest_level: Nesting level of the array column.
    :returns: Nested transform function.
    """
    if nest_level <= 0:
        # If the nesting level is 0 or less, we have run out of arrays and are now at
        # the scalar level.
        return func
    return lambda x: F.transform(x, nested_transform(func, nest_level=nest_level - 1))


def nested_lambda(func: Callable, nest_level: int) -> Callable:
    """
    Similar to nested_transform, however in this case creates a lambda function that
    applies a function to each element of a nested array. Used within UDFs to apply a
    function to each element of a nested array.

    :param func: Function to apply to each element of the nested array column.
    :param nest_level: Nesting level of the array column.
    :returns: Nested lambda function.
    """

    def apply_func_to_list(x, function):
        return [function(y) for y in x]

    if nest_level == 0:
        return func
    elif nest_level < 0:
        raise ValueError("nest_level must be greater than or equal to 0.")
    return lambda x: apply_func_to_list(
        x, nested_lambda(func, nest_level=nest_level - 1)
    )


def build_udf_return_type(element_dtype: DataType, nest_level: int) -> DataType:
    """
    Builds the return type of a UDF that applies a function to each element of a nested
    array. This is used to specify the return type of the UDF.

    :param element_dtype: DataType of the elements of the nested array.
    :param nest_level: Nesting level of the array column.
    :returns: DataType of the return type of the UDF.
    """
    if nest_level == 0:
        return element_dtype
    elif nest_level < 0:
        raise ValueError("nest_level must be greater than or equal to 0.")
    for _ in range(nest_level):
        element_dtype = ArrayType(element_dtype)
    return element_dtype


def flatten_nested_arrays(column: Column, column_data_type: DataType) -> Column:
    """
    Flattens a nested array to a single array.

    :param column: Spark array column to be flattened
    :param column_data_type: Datatype of the Spark array column.
    :returns: Flattened array column.
    """
    nested_level = get_array_nesting_level(column_dtype=column_data_type)
    flattened_column = column
    for _ in range(1, nested_level):
        flattened_column = F.flatten(flattened_column)

    return flattened_column



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import (
    MultiInputSingleOutputParams,
    SingleInputSingleOutputParams,
    StringRegexParams,
)
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringReplaceLayer

from .base import BaseTransformer


class StringReplaceParams(StringRegexParams):
    """
    Mixin class containing StringMatchConstant, StringReplaceConstant needed for string
    replace layers.
    """

    stringMatchConstant = Param(
        Params._dummy(),
        "stringMatchConstant",
        "String match constant to use in string replace",
        typeConverter=TypeConverters.toString,
    )

    stringReplaceConstant = Param(
        Params._dummy(),
        "stringReplaceConstant",
        "String replace constant to use in string replace",
        typeConverter=TypeConverters.toString,
    )

    def setStringMatchConstant(self, value: str) -> "StringReplaceParams":
        """
        Sets the stringConstant parameter.

        :param value: String constant value to use in different string transformers.
        :returns: Instance of class mixed in.
        """
        return self._set(stringMatchConstant=value)

    def getStringMatchConstant(self) -> str:
        """
        Gets the stringConstant parameter.

        :returns: String replacement constant value to use in string replace layer.
        """
        return self.getOrDefault(self.stringMatchConstant)

    def setStringReplaceConstant(self, value: str) -> "StringReplaceParams":
        """
        Sets the stringReplaceConstant parameter.
        """
        return self._set(stringReplaceConstant=value)

    def getStringReplaceConstant(self) -> str:
        """
        Gets the stringReplaceConstant parameter.

        :returns: String replacement constant value to use in string replace layer.
        """
        return self.getOrDefault(self.stringReplaceConstant)


class StringReplaceTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    StringReplaceParams,
):
    """
    String replace Spark Transformer for use in Spark Pipelines.
    This transformer performs a string replace operation on the input column.

    The transformer takes up to 3 input columns.
    The first input column is always required and is the column we operate on.
    A match constant/column is provided to match against and the replace constant/column
    is provided to replace the matched substrings with.

    KNOWN ISSUE: when replacing with a string that contains a backslash,
    the backslash must be double escaped (\\\\) in order to be added properly.
    This is consistent in both spark and tensorflow components.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        stringMatchConstant: Optional[str] = None,
        stringReplaceConstant: Optional[str] = None,
        regex: bool = False,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes an StringReplaceTransformer transformer.

        :param inputCol: Input column name.
        :param inputCols: List of up to 3 input column names.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param stringMatchConstant: String constant to match for.
        operation.
        :param stringReplaceConstant: String constant to replace with.
        :param regex: Whether to allow regex-matching in the string matching.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            regex=False, stringMatchConstant=None, stringReplaceConstant=None
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def setInputCols(self, value: List[str]) -> "StringReplaceTransformer":
        """
        Overrides setting the input columns for the transformer.
        Throws an error if we have more than 3 input columns.

        :param value: List of input columns.
        :returns: Class instance.
        """
        if len(value) > 3:
            raise ValueError(
                """When setting inputCols for StringReplaceTransformer,
                there must be 1-3 columns."""
            )

        return self._set(inputCols=value)

    def _construct_input_cols(self) -> List[Column]:
        """
        Constructs the input columns for the transformer.

        :returns: List of pyspark columns.
        """
        if self.isDefined("inputCol"):
            input_cols = [F.col(self.getInputCol())]
        elif self.isDefined("inputCols"):
            input_cols = [F.col(c) for c in self.getInputCols()]
        else:
            raise ValueError("Must specify either inputCol or inputCols.")

        if self.getStringReplaceConstant() is not None:
            input_cols.insert(
                len(input_cols),
                F.lit(self.getStringReplaceConstant()).alias(
                    self.uid + "_stringReplaceConstant"
                ),
            )
        if self.getStringMatchConstant() is not None:
            input_cols.insert(
                1,
                F.lit(self.getStringMatchConstant()).alias(
                    self.uid + "_stringMatchConstant"
                ),
            )

        if len(input_cols) > 3:
            raise ValueError(
                """When setting inputCols for StringReplaceTransformer,
            there must be 1-3 columns. In the case 3 columns are specified, string
            constants should not be specified."""
            )
        elif len(input_cols) < 3:
            raise ValueError(
                """Less than 3 input columns were provided,
            but no string constants were passed as arguments."""
            )

        return input_cols

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which contains the result of the string replace operation.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self._construct_input_cols()
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]
        regex = self.getRegex()

        def string_replace(
            x: Column, input_col_names: List[str], regex: bool
        ) -> Column:
            col_expr = (
                F.regexp_replace(
                    x[input_col_names[0]],
                    F.regexp_replace(
                        x[input_col_names[1]], "[^A-Za-z0-9]", "\\\\" + "$0"
                    ),
                    x[input_col_names[2]],
                )
                if not regex
                else F.regexp_replace(
                    x[input_col_names[0]],
                    F.when(x[input_col_names[1]] == F.lit(""), F.lit("^$")).otherwise(
                        x[input_col_names[1]]
                    ),
                    x[input_col_names[2]],
                )
            )
            return col_expr

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: string_replace(x, input_col_names, regex),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the StringReplaceLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a string replace operation.
        """
        return StringReplaceLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            regex=self.getRegex(),
            string_match_constant=self.getStringMatchConstant(),
            string_replace_constant=self.getStringReplaceConstant(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input, map_fn_w_axis

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class OrdinalArrayEncodeLayer(BaseLayer):
    """
    Transformer that encodes an array of strings into an array of integers.

    The transformer will map each unique string in the array to an integer,
    according to the order in which they appear in the array. It will also
    ignore the pad value if specified.
    """

    def __init__(
        self,
        pad_value: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        axis: int = -1,
        name: str = None,
        **kwargs,
    ) -> None:
        """
        Initializes the OrdinalArrayEncodeLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param pad_value: The value which pad the array and as a result should be
        ignored in the encoding process.

        :returns: None
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.pad_value = pad_value
        self.axis = axis

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the ordinal encoding on the input dataset.
        Example:
         input_tensor = tf.Tensor([
            ['a', 'a', 'a', 'b', 'c', '-1', '-1', '-1'],
            ['x', 'x', 'x', 'x', 'y', 'z', '-1', '-1'],
            ]
         )

        Output: tf.Tensor([[
            [0, 0, 0, 1, 2, -1, -1, -1],
            [0, 0, 0, 0, 1, 2, -1, -1],
            ]
        )

        :param inputs: The input tensor.
        :returns: Transformed tensor.
        """

        @tf.function
        def _transform_row(input_row: Tensor) -> Tensor:
            if self.pad_value is None:
                converted_tensor = tf.unique(input_row).idx
            else:
                not_pad_mask = tf.where(
                    tf.not_equal(input_row, self.pad_value),
                    tf.constant(True),
                    tf.constant(False),
                )
                # If all values are the pad value return -1s
                if not tf.reduce_any(not_pad_mask):
                    converted_tensor = tf.fill(tf.shape(input_row), -1)
                else:
                    non_pad_values = tf.boolean_mask(input_row, not_pad_mask)
                    first_non_pad_value = non_pad_values[0]
                    replace_pad_with_first = tf.where(
                        tf.equal(input_row, self.pad_value),
                        first_non_pad_value,
                        input_row,
                    )
                    converted_tensor = tf.where(
                        not_pad_mask,
                        tf.unique(replace_pad_with_first).idx,
                        tf.constant(-1),
                    )
            return self._cast(converted_tensor, cast_dtype=tf.int32.name)

        output = map_fn_w_axis(
            elems=inputs,
            fn=_transform_row,
            axis=self.axis,
            fn_output_signature=tf.int32,
        )

        return output

    def get_config(self):
        """
        Gets the configuration of the OrdinalArrayEncoder layer.
        Used for saving and loading from a model.

        Specifically adds the `pad_value` value to the configuration.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"pad_value": self.pad_value, "axis": self.axis})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringReplaceLayer(BaseLayer):
    """
    StringReplaceLayer layer for TensorFlow.
    """

    def __init__(
        self,
        string_match_constant: Optional[str] = None,
        string_replace_constant: Optional[str] = None,
        regex: bool = False,
        name: Optional[str] = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the StringReplaceLayer layer.

        WARNING: While it works, the use of tensors in matching/replacement
        is not recommended due to the complexity of the regex matching which requires
        use of a map_fn. This will be comparatively VERY slow and may not be suitable
        for inference use-cases.
        If you know where in the string the match is, you will be much
        better off slicing the string and checking for equality.

        :param string_match_constant: The string to match against and replace.
            Defaults to `None`.
        :param string_replace_constant: The string to replace the matched string with.
            Defaults to `None`.
        :param regex: Whether to treat the string match as a regular expression.
            Defaults to `False`. In the case regex is enabled, the string_match_constant
            or second input tensor elements are treated as a regex pattern. Please be
            aware that while testing has tried to catch corner cases, this is not
            guaranteed to be bug-free due to slight differences in the regex
            implementations between Spark and TensorFlow.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.string_match_constant = string_match_constant
        self.string_replace_constant = string_replace_constant
        self.regex = regex

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs) -> Tensor:
        """
        Checks for the existence of a substring/pattern within a tensor and replaces
        if there is a match.

        KNOWN ISSUE: when replacing with a string that contains a backslash,
        the backslash must be double escaped (\\\\) in order to be added properly.
        This is consistent in both spark and tensorflow components.

        WARNING: While it works, the use of tensors in matching/replacement
        is not recommended due to the complexity of the regex matching which requires
        use of a map_fn. This will be comparatively VERY slow and may not be suitable
        for inference use-cases.
        If you know where in the string the match is, you will be much
        better off slicing the string and checking for equality.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param: inputs: A string tensor or iterable of up to three string
            tensors.
            In the case multiple tensors are passed, require that the order of inputs is
             [string input, {string match tensor}, {string replace tensor}].
        :returns: A string tensor of regex replaced strings.
        """

        match_all_pattern = r"([\w]\\+\_+\!+\?+)*"

        # Case both match and replacement are constant
        if (
            self.string_replace_constant is not None
            and self.string_match_constant is not None
        ):
            if len(inputs) == 1:
                # Need the tensor for shapes to be consistent
                input_tensor = inputs[0]

                match_substring = self.string_match_constant

                if not self.regex:
                    match_substring = self._escape_special_characters(match_substring)

                # Calls regex replace function on the input tensor, matching
                # with match constant and replacing with replace constant
                replaced_tensor = tf.strings.regex_replace(
                    input_tensor,
                    tf.constant(
                        match_all_pattern + match_substring + match_all_pattern
                        if match_substring != ""
                        else "^$"
                    ),
                    tf.constant(self.string_replace_constant),
                )

            else:
                raise ValueError(
                    """When string_match_constant and string_replace_constant are
                    defined, expected a single tensor as input."""
                )
        else:
            # Preserve input shape
            input_shape = tf.shape(inputs[0])
            # Generate a tensor that can be used by map_fn
            # First we define 3 tensors, the input string, the match string and the
            # replace string
            string_tensor = inputs[0]
            match_substring = (
                tf.constant(self.string_match_constant, shape=string_tensor.shape)
                if self.string_match_constant is not None
                else inputs[1]
            )
            replace_substring = (
                tf.constant(self.string_replace_constant, shape=string_tensor.shape)
                if self.string_replace_constant is not None
                else inputs[1 + (len(inputs) == 3)]
            )

            # Stack the input, match and replace elements into a single tensor
            # then flatten for use in map_fn
            mappable_tensor = tf.stack(
                [string_tensor, match_substring, replace_substring], axis=-1
            )
            mappable_tensor = tf.reshape(mappable_tensor, [-1, 3])

            def _tensor_replace(x):
                match_substring = x[1]
                if not self.regex:
                    match_substring = self._escape_special_characters(x[1])
                return tf.strings.regex_replace(
                    input=x[0],
                    pattern=match_all_pattern + match_substring + match_all_pattern
                    if match_substring != ""
                    else "^$",
                    rewrite=x[2],
                )

            # TODO: tf.vectorized_map may be slightly faster with larger batches
            #  but this requires some refactoring
            replaced_tensor = tf.map_fn(
                _tensor_replace,
                elems=mappable_tensor,
                dtype=tf.string,
            )

            # Reshape to the preserved input shape
            replaced_tensor = tf.reshape(replaced_tensor, input_shape)

        return replaced_tensor

    def _escape_special_characters(
        self, string_to_escape: Union[str, Tensor]
    ) -> Union[str, Tensor]:
        """
        Escapes special characters in a string so they are not parsed as regex.
        :param string: The string or string tensor to escape special characters in.
        :returns: The escaped string or string tensor.
        """

        for char in [
            ".",
            "^",
            "$",
            "*",
            "+",
            "?",
            "{",
            "}",
            "[",
            "]",
            "(",
            ")",
            "|",
        ]:
            if isinstance(string_to_escape, str):
                string_to_escape = string_to_escape.replace(char, "\\\\" + char)
            else:
                string_to_escape = tf.strings.regex_replace(
                    string_to_escape, "\\" + char, "\\\\" + char
                )
        return string_to_escape

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringReplace layer.
        Used for saving and loading the layer from disk.

        Specifically, `regex`, `string_match_constant` and `string_replace_constant`
        are added to the config.

        :returns: Dictionary configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "regex": self.regex,
                "string_match_constant": self.string_match_constant,
                "string_replace_constant": self.string_replace_constant,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql.types import DataType, DoubleType, LongType

from kamae.spark.params import (
    DateTimeParams,
    SingleInputSingleOutputParams,
    UnixTimestampParams,
)
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import UnixTimestampToDateTimeLayer


class UnixTimestampToDateTimeTransformer(
    BaseTransformer, SingleInputSingleOutputParams, UnixTimestampParams, DateTimeParams
):
    """
    Transformer that converts a unix timestamp to a datetime.

    The unix timestamp can be in milliseconds or seconds, set by the `unit` parameter.
    If the `includeTime` parameter is set to True, the output will be in
    yyyy-MM-dd HH:mm:ss.SSS format. If set to False, the output will be in
    yyyy-MM-dd format.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        unit: str = "s",
        includeTime: bool = True,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialises the UnixTimestampToDateTimeTransformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param unit: Unit of the timestamp. Can be `milliseconds` (shorthand `ms`)
         or `seconds` (shorthand `s`). Default is `s` (seconds).
        :param includeTime: Whether to include the time in the output. Default is True.
        :param layerName: Layer name. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(unit="s", includeTime=True)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)
        # TODO: Remove this when we only support PySpark 3.5+. It is only used to get
        #  the timezone set by the user for datetime operations. In 3.5+ we can use the
        #  current_timezone() function. Also is there a better way to access this than
        #  inside a class attribute? Setting it at the top of the file causes issues
        #  in tests as we import all transformers when the package is loaded.
        self.spark = SparkSession.builder.getOrCreate()

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            DoubleType(),
            LongType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input integer timestamp to the date string with format
        yyyy-MM-dd HH:mm:ss.SSS.

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def unix_timestamp_to_datetime(
            unix_timestamp: Column, include_time: bool
        ) -> Column:
            """
            Returns the date in yyyy-MM-dd HH:mm:ss.SSS format from a Unix timestamp
            in seconds.

            :param unix_timestamp: Unix timestamp in seconds.
            :param include_time: Whether to include the time in the output.
            :returns: Column of the date in yyyy-MM-dd HH:mm:ss.SSS format if
            include_time is True, otherwise in yyyy-MM-dd format.
            """
            # from_unixtime throws away milliseconds, so we have to calculate them
            # separately
            milliseconds_3dp = (
                (unix_timestamp - F.floor(unix_timestamp)) * 1000.0
            ).cast("int")
            local_datetime_str_wo_millis = F.from_unixtime(
                unix_timestamp, format="yyyy-MM-dd HH:mm:ss"
            )
            local_datetime_str_w_millis = F.concat(
                local_datetime_str_wo_millis,
                F.lit("."),
                F.lpad(milliseconds_3dp.cast("string"), 3, "0"),
            )
            local_datetime = F.to_timestamp(
                local_datetime_str_w_millis, format="yyyy-MM-dd HH:mm:ss.SSS"
            )
            utc_datetime = F.to_utc_timestamp(
                local_datetime, self.spark.conf.get("spark.sql.session.timeZone")
            )
            date_fmt = "yyyy-MM-dd HH:mm:ss.SSS" if include_time else "yyyy-MM-dd"
            return F.date_format(utc_datetime, date_fmt)

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: unix_timestamp_to_datetime(x, self.getIncludeTime())
            if self.getUnit() == "s"
            else unix_timestamp_to_datetime(x / F.lit(1000), self.getIncludeTime()),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the unix timestamp to date transform.

        :returns: Tensorflow layer that performs the unix timestamp to date transform.
        """
        return UnixTimestampToDateTimeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            unit=self.getUnit(),
            include_time=self.getIncludeTime(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from bisect import bisect_right
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType, IntegerType, LongType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils.transform_utils import (
    single_input_single_output_scalar_udf_transform,
)
from kamae.tensorflow.layers import BucketizeLayer

from .base import BaseTransformer


class BucketizeParams(Params):
    """
    Mixin class containing splits parameter needed for bucketing.
    """

    splits = Param(
        Params._dummy(),
        "splits",
        "List of split points for bucketing.",
        typeConverter=TypeConverters.toListFloat,
    )

    @staticmethod
    def check_splits_sorted(splits: List[float]):
        """
        Checks that the splits parameter is sorted.

        :param splits: List of float values to use for bucketing.
        """
        if splits is not None and splits != sorted(splits):
            raise ValueError("`splits` argument must be a sorted list!")

    def setSplits(self, value: List[float]) -> "BucketizeParams":
        """
        Sets the splits parameter.

        :param value: List of float values to use for bucketing.
        :returns: Instance of class mixed in.
        """
        self.check_splits_sorted(value)
        return self._set(splits=value)

    def getSplits(self) -> List[float]:
        """
        Gets the splits parameter.

        :returns: List of float values to use for bucketing.
        """
        return self.getOrDefault(self.splits)


class BucketizeTransformer(
    BaseTransformer,
    BucketizeParams,
    SingleInputSingleOutputParams,
):
    """
    BucketizeLayer Spark Transformer for use in Spark pipelines.
    This transformer buckets a numerical column into bins.
    Buckets will be created based on the splits parameter.
    The bins are integer values starting at 1 and ending at the number of splits + 1.
    The 0 index is reserved for masking/padding.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        splits: Optional[List[float]] = None,
    ) -> None:
        """
        Initializes an BucketizeTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param splits: List of float values to use for bucketing.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [IntegerType(), LongType(), FloatType(), DoubleType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the `inputCol` bucketed into bins accoring to the `splits` parameter.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        splits = self.getSplits()
        # We need to create a UDF to perform binary search on the splits.

        def bucketize(value):
            # If null, keep null. There is no best bucket to place these into.
            if value is None:
                return None
            # We add 1 because we want to reserve the 0 index for mask/padding.
            return bisect_right(splits, value) + 1

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_udf_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: bucketize(x),
            udf_return_element_datatype=IntegerType(),
        )

        return dataset.withColumn(
            self.getOutputCol(),
            output_col,
        )

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the BucketizeLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a bucketing operation.
        """
        return BucketizeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            splits=self.getSplits(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import BaseEstimator  # noqa: F401
from .conditional_standard_scale import ConditionalStandardScaleEstimator  # noqa: F401
from .impute import ImputeEstimator  # noqa: F401
from .one_hot_encode import OneHotEncodeEstimator  # noqa: F401
from .shared_one_hot_encode import SharedOneHotEncodeEstimator  # noqa: F401
from .shared_string_index import SharedStringIndexEstimator  # noqa: F401
from .single_feature_array_standard_scale import (  # noqa: F401
    SingleFeatureArrayStandardScaleEstimator,
)
from .standard_scale import StandardScaleEstimator  # noqa: F401
from .string_index import StringIndexEstimator  # noqa: F401



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional


class LayerNameMixin:
    """
    Mixin class for a layer name.
    """

    _layer_name: Optional[str]

    @property
    def layer_name(self) -> str:
        """
        Gets the layer name.

        :returns: String of layer name.
        """
        return self._layer_name



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
from farmhash import hash64
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import ArrayType, DataType, StringType

from kamae.spark.utils import flatten_nested_arrays


def safe_hash64(label: str) -> int:
    """
    Attempt to hash a string label and raise an exception if it contains a null
    character.

    :param label: String to hash.
    :raises ValueError: If the label contains a null character.
    :returns: Hashed integer value.
    """
    try:
        return hash64(label)
    except ValueError as e:
        if str(e) == "embedded null character":
            raise ValueError(
                f"""Label {label} contains a null character.
                These cause issues with hashing. You should remove these from your data.
                https://en.wikipedia.org/wiki/Null_character
                """
            ).with_traceback(e.__traceback__)
        else:
            raise e


def collect_labels_array(
    dataset: DataFrame,
    column: Column,
    column_datatype: DataType,
    string_order_type: str = "frequencyDesc",
    mask_token: Optional[str] = None,
    max_num_labels: Optional[int] = None,
) -> List[str]:
    """
    Collects the string labels from a given column in a given dataset.

    :param dataset: Spark input dataset.
    :param column: Spark input column.
    :param column_datatype: Spark input column datatype.
    :param string_order_type: String order type, one of
    'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc', 'alphabeticalDesc'
    :param mask_token: Optional value to mask when indexing.
    :param max_num_labels: Optional value to limit the number of labels.

    :returns: List of string labels.
    """
    possible_order_options = [
        "frequencyAsc",
        "frequencyDesc",
        "alphabeticalAsc",
        "alphabeticalDesc",
    ]
    if string_order_type not in possible_order_options:
        raise ValueError(
            f"string_order_type must be one of {', '.join(possible_order_options)}"
        )

    if (max_num_labels is not None) and (not isinstance(max_num_labels, int)):
        raise ValueError("max_labels_count must be an integer")

    input_col_an_array = isinstance(column_datatype, ArrayType)

    if input_col_an_array:
        # Flatten the array to a single array
        flattened_array_col = flatten_nested_arrays(
            column=column, column_data_type=column_datatype
        )
        # If the input column is an array, we need to flatten it twice
        # before we can collect the string labels
        rdd_vals = (
            dataset.select(flattened_array_col.cast("array<string>"))
            .rdd.map(lambda x: x.asDict().values())
            .flatMap(lambda x: x)
            .flatMap(lambda x: x)
        )
    else:
        # Otherwise, since it is scalar we can just collect the string labels
        rdd_vals = dataset.select(column.cast("string")).rdd.map(lambda x: x[0])

    # If the mask token is defined, we remove it from the labels.
    # This is because otherwise Tensorflow can throw an error when
    # the mask token is not in the correct position of the labels.
    rdd_vals = (
        rdd_vals.filter(lambda x: x != mask_token)
        if mask_token is not None
        else rdd_vals
    )
    sort_ascending = string_order_type in ["alphabeticalAsc", "frequencyAsc"]
    if string_order_type in ["frequencyAsc", "frequencyDesc"]:
        label_vals = (
            rdd_vals.filter(lambda x: x is not None)
            .map(lambda x: (x, 1))
            .reduceByKey(lambda x, y: x + y)
            .sortBy(lambda x: x[1], ascending=sort_ascending)
            .map(lambda x: x[0])
        )
    else:
        label_vals = (
            rdd_vals.filter(lambda x: x is not None)
            .distinct()
            .sortBy(lambda x: x, ascending=sort_ascending)
        )

    if max_num_labels is not None:
        return label_vals.take(max_num_labels)
    else:
        return label_vals.collect()


def collect_labels_array_from_multiple_columns(
    dataset: DataFrame,
    columns: List[Column],
    column_datatypes: List[DataType],
    string_order_type: str = "frequencyDesc",
    mask_token: Optional[str] = None,
    max_num_labels: Optional[int] = None,
) -> List[str]:
    """
    Collects the string labels across multiple columns in a given dataset.

    :param dataset: Spark input dataset
    :param columns: List of Spark input columns.
    :param column_datatypes: List of Spark input column datatypes.
    :param string_order_type: String order type, one of
    'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc', 'alphabeticalDesc'
    :param mask_token: Optional value to mask when indexing.
    :param max_num_labels: Optional value to limit the number of labels.
    :returns: List of string labels.
    """
    flattened_arrays = []
    for column, column_datatype in zip(columns, column_datatypes):
        column_is_array = isinstance(column_datatype, ArrayType)
        if column_is_array:
            # Flatten any arrays and add to the list
            flattened_arrays.append(
                flatten_nested_arrays(column=column, column_data_type=column_datatype)
            )
        else:
            flattened_arrays.append(F.array(column))

    concatenated_array = F.concat(*flattened_arrays).cast("array<string>")

    return collect_labels_array(
        dataset=dataset,
        column=concatenated_array,
        column_datatype=ArrayType(StringType()),
        string_order_type=string_order_type,
        mask_token=mask_token,
        max_num_labels=max_num_labels,
    )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import BooleanType, DataType, FloatType, IntegerType, StringType

from kamae.spark.params import PadValueParams, SingleInputSingleOutputParams
from kamae.spark.utils import (
    get_array_nesting_level_and_element_dtype,
    single_input_single_output_array_transform,
)
from kamae.tensorflow.layers import ArrayCropLayer

from .base import BaseTransformer


class ArrayCropParams(PadValueParams):
    """
    Mixin class containing pad value parameters needed
    for array crop transformers.
    """

    arrayLength = Param(
        PadValueParams._dummy(),
        "arrayLength",
        "The length to crop or pad the arrays to. Defaults to 128.",
        typeConverter=TypeConverters.toInt,
    )

    def setArrayLength(self, value: int) -> "ArrayCropParams":
        """
        Sets the parameter array length to the given value.
        :param value: array length.
        :returns: Instance of class mixed in.
        """
        if value < 1:
            raise ValueError("Array length must be greater than 0.")
        return self._set(arrayLength=value)

    def getArrayLength(self) -> int:
        """
        Gets the array length parameter.
        :returns: array length.
        """
        return self.getOrDefault(self.arrayLength)


class ArrayCropTransformer(
    BaseTransformer, SingleInputSingleOutputParams, ArrayCropParams
):
    """
    Transformer that reshapes arrays into consistent shapes by
    either cropping or padding.

    If the tensor is shorter than the specified length, it is
    padded with specified pad value.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[Union[str, int, float]] = None,
        outputDtype: Optional[Union[str, int, float]] = None,
        layerName: Optional[str] = None,
        arrayLength: Optional[int] = 128,
        padValue: Optional[Union[str, int, float]] = None,
    ) -> None:
        """
        Initialises the ArrayCropTransformer
        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        :param arrayLength: The length to crop or pad the arrays to. Defaults to 128.
        :param padValue: The value pad the arrays with. Defaults to `None`.
        :returns: None
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)
        self._pad_type_to_valid_element_types = {
            "int": ["int", "bigint", "smallint"],
            "float": ["float", "double", "decimal(10,0)"],
            "string": ["string"],
            "boolean": ["boolean"],
        }

    @staticmethod
    def _get_pad_value_type(pad_value) -> Optional[DataType]:
        if isinstance(pad_value, int):
            return IntegerType()
        if isinstance(pad_value, str):
            return StringType()
        if isinstance(pad_value, float):
            return FloatType()
        if isinstance(pad_value, bool):
            return BooleanType()
        raise TypeError(f"Unsupported pad value type: {type(pad_value)}")

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Performs the cropping and/or padding on the input dataset.
        Example, crop to length 3, with value '-1':

         dataset = spark.Dataframe(
            [
                ['a', 'a', 'a', 'b', 'c'],
                ['x', 'z', 'y'],
                ['a', 'b',],
                ['a', 'x', 'a', 'b',],
                []
            ],
            'input_col'
         )
         Output: spark.Dataframe(
            [
                ['a', 'a', 'a', 'b', 'c'],
                ['x', 'z', 'y'],
                ['a', 'b',],
                ['a', 'x', 'a', 'b',],
                []
            ],
            [
                ['a', 'a', 'a'],
                ['x', 'z', 'y'],
                ['a', 'b', '-1'],
                ['a', 'x', 'a'],
                ['-1', '-1', '-1']
            ],
            'input_col', 'output_col'
        )
        :param dataset: The input dataframe.
        :returns: Transformed pyspark dataframe.
        """
        pad_value_spark_type = self._get_pad_value_type(self.getPadValue())
        input_col_type = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        input_col_element_type = get_array_nesting_level_and_element_dtype(
            input_col_type
        )[1]

        if (
            input_col_element_type.simpleString()
            not in self._pad_type_to_valid_element_types[
                pad_value_spark_type.simpleString()
            ]
        ):
            raise ValueError(
                f"""
            The pad value type '{type(pad_value_spark_type)}' does
            not match the element type of the input
            column '{type(input_col_element_type)}'.
            """
            )

        output_col = single_input_single_output_array_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_col_type,
            func=lambda x: F.concat(
                F.slice(x, 1, self.getArrayLength()),
                F.array_repeat(
                    F.lit(self.getPadValue()),
                    self.getArrayLength() - F.size(x),
                ),
            ),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the array cropping and padding.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the array cropping and padding operation.
        """
        return ArrayCropLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            array_length=self.getArrayLength(),
            pad_value=self.getPadValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class LogicalOrLayer(BaseLayer):
    """
    Performs the or(x, y) operation on a given input tensor.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initializes the LogicalOrLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bool]

    @enforce_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Performs the or(x, y) operation on an iterable of input tensors

        Decorated with `@enforce_multiple_tensor_input` to ensure that the input
        is an iterable of tensors. Raises an error if a single tensor is passed
        in.

        :param inputs: Iterable of tensors to perform the or(x, y) operation on.
        :returns: The tensor resulting from the or(x, y) operation.
        """
        if len(inputs) == 1:
            raise ValueError("Expected multiple inputs, but got a single input")
        return reduce(tf.math.logical_or, inputs)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the LogicalOr layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, IntegerType, StringType

from kamae.spark.params import PadValueParams, SingleInputSingleOutputParams
from kamae.spark.utils import (
    ordinal_array_encode_udf,
    single_input_single_output_array_udf_transform,
)
from kamae.tensorflow.layers import OrdinalArrayEncodeLayer

from .base import BaseTransformer


class OrdinalArrayEncodeTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    PadValueParams,
):
    """
    Transformer that encodes an array of strings into an array of integers.

    The transformer will map each unique string in the array to an integer,
    according to the order in which they appear in the array. It will also
    ignore the pad value if specified.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        padValue: Optional[str] = None,
    ) -> None:
        """
        Initialises the OrdinalArrayEncodeTransformer
        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        :param padValue: The value to be considered as padding. Defaults to `None`.
        :returns: None
        """
        super().__init__()
        self._setDefault(padValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            StringType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Performs the ordinal encoding on the input dataset.
        Example:
         dataset = spark.Dataframe(
            [
                ['a', 'a', 'a', 'b', 'c', '-1', '-1', '-1'],
                ['x', 'x', 'x', 'x', 'y', 'z', '-1', '-1'],
            ],
            'input_col'
         )
         Output: spark.Dataframe(
            [
                ['a', 'a', 'a', 'b', 'c', '-1', '-1', '-1'],
                ['x', 'x', 'x', 'x', 'y', 'z', '-1', '-1'],
            ],
            [
                [0, 0, 0, 1, 2, -1, -1, -1],
                [0, 0, 0, 0, 1, 2, -1, -1],
            ],
            'input_col', 'output_col'
        )
        :param dataset: The input dataframe.
        :returns: Transformed pyspark dataframe.
        """
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        if not isinstance(input_datatype, ArrayType):
            raise ValueError(
                f"""Input column {self.getInputCol()} must be of ArrayType.
                        Got {input_datatype} instead."""
            )

        output_col = single_input_single_output_array_udf_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: ordinal_array_encode_udf(x, self.getPadValue()),
            udf_return_element_datatype=IntegerType(),
        )

        return dataset.withColumn(
            self.getOutputCol(),
            output_col,
        )

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the ordinal array encoding.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the ordinal array encoding operation.
        """
        return OrdinalArrayEncodeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            pad_value=self.getPadValue(),
            axis=-1,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import CurrentDateTimeLayer


class CurrentDateTimeTransformer(BaseTransformer, SingleInputSingleOutputParams):
    """
    Returns the current UTC datetime in yyyy-MM-dd HH:mm:ss.SSS format.

    NOTE: Parity between this and its TensorFlow counterpart is very difficult at the
    millisecond level. We have to round the TensorFlow timestamp to the 3rd decimal
    place for milliseconds, because  Spark already truncates to 3 decimal places.
    Therefore, parity is not guaranteed at this precision.

    It is recommended not to rely on parity at the millisecond level.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialises the CurrentDateTime layer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)
        # TODO: Remove this when we only support PySpark 3.5+. It is only used to get
        #  the timezone set by the user for datetime operations. In 3.5+ we can use the
        #  current_timezone() function. Also is there a better way to access this than
        #  inside a class attribute? Setting it at the top of the file causes issues
        #  in tests as we import all transformers when the package is loaded.
        self.spark = SparkSession.builder.getOrCreate()

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Returns a column of the current datetime. If an array column is provided,
        we return an array column of identical structure with elements populated by
        the current datetime

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def current_utc_timestamp() -> Column:
            """
            Returns the current UTC timestamp. Spark respects the timezone
            set in the Spark session for datetime operations, so we need to be agnostic
            to the timezone set by the user.

            :returns: Column of the current UTC timestamp.
            """
            local_timestamp = F.localtimestamp()
            # TODO: Replace this with current_timezone() once we only support PySpark
            #  3.5+
            local_timezone = self.spark.conf.get("spark.sql.session.timeZone")
            return F.date_format(
                F.to_utc_timestamp(local_timestamp, local_timezone),
                "yyyy-MM-dd HH:mm:ss.SSS",
            )

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: current_utc_timestamp(),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer.

        :returns: CurrentDateTimeLayer Tensorflow layer.
        """
        return CurrentDateTimeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from random import choice
from string import ascii_uppercase
from typing import List, Optional, Tuple

import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, NumericType

from kamae.spark.params import (
    HasInputDtype,
    HasLayerName,
    HasOutputDtype,
    InputOutputExtractor,
)
from kamae.spark.utils import (
    get_element_type,
    single_input_single_output_scalar_transform,
)


class SparkOperation(
    ABC, HasLayerName, HasInputDtype, HasOutputDtype, InputOutputExtractor
):
    """
    Abstract class used in Spark transformers and estimators. Provides common utils for
    param setting, input/output dtype casting, and layer name setting.
    """

    def __init__(self):
        """
        Initializes the spark operation class.
        """
        super().__init__()
        self._setDefault(layerName=self.uid, inputDtype=None, outputDtype=None)
        self.tmp_column_suffix = self.generate_tmp_column_suffix()

    @property
    @abstractmethod
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the spark operation.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the spark operation.
        """
        raise NotImplementedError

    @keyword_only
    def setParams(self, **kwargs) -> "SparkOperation":
        """
        Sets all given keyword parameters.

        :returns: Class instance.
        """
        input_kwargs = self._input_kwargs
        if (
            "inputCol" in input_kwargs
            and "inputCols" in input_kwargs
            and input_kwargs["inputCol"] is not None
            and input_kwargs["inputCols"] is not None
        ):
            raise ValueError("Only one of inputCol or inputCols can be set, not both.")
        for param_name, param_value in input_kwargs.items():
            # Only if the param value is not None, do we set the param
            if param_value is not None:
                # Get the setter method for the parameter
                setter_method_name = f"set{param_name[0].upper()}{param_name[1:]}"
                setter_method = getattr(self, setter_method_name)
                # Set the parameter
                setter_method(param_value)
        return self

    @staticmethod
    def generate_tmp_column_suffix(str_len: int = 25):
        """
        Returns a random string of length `str_len` to append to temporary columns.

        This ensures that there is minimal collision between temporary columns created
        and the original columns.

        :param str_len: Length of the random string.
        :returns: Random string of length `str_len`.
        """
        return "".join(choice(ascii_uppercase) for _ in range(str_len))

    @staticmethod
    def _resolve_tmp_from_true_column_name(
        column_name: str, suffix: str, reverse: bool = False
    ):
        """
        Resolves the temporary column name from the true column name or vice versa.

        Used to rename input columns when casting is necessary to avoid overwriting
        the original columns.

        :param column_name: The current column name.
        :param suffix: The suffix to append to the column name.
        :param reverse: Whether to resolve the temporary column name from the true
        column name or vice versa.
        """
        return (
            f"{column_name}_{suffix}"
            if not reverse
            else column_name.replace(f"_{suffix}", "")
        )

    @staticmethod
    def _cast(column: Column, column_dtype: DataType, cast_dtype: str) -> Column:
        """
        Casts the input column to the specified datatype.

        :param column: Pyspark column to cast.
        :param column_dtype: The datatype class of the column.
        :param cast_dtype: The datatype string to cast the column to.
        :returns: Pyspark column cast to the specified datatype.
        """
        # There is an edge case where we can have a negatively signed 0.
        # This will not match tensorflow, so we will ensure the sign of zeros is
        # always positive.
        if isinstance(column_dtype, NumericType):
            # We need to cast back to the original type after multiplying by the sign
            # since the multiplication by sign will return a double. Therefore, we can
            # have cases where the original type is an int, it is cast to a double, and
            # then we want a string. So we have 1 -> 1.0 -> "1.0" rather than "1".
            # TODO: I really don't like this, there must be a better way to do this.
            func = lambda x: F.when(  # noqa: E731
                x == F.lit(0),
                (x * F.signum(x)).cast(column_dtype.simpleString()).cast(cast_dtype),
            ).otherwise(x.cast(cast_dtype))
        else:
            func = lambda x: x.cast(cast_dtype)  # noqa: E731
        return single_input_single_output_scalar_transform(
            input_col=column, input_col_datatype=column_dtype, func=func
        )

    def _cast_input_output_columns(
        self, columns: List[Column], column_datatypes: List[DataType], ingress: bool
    ) -> List[Tuple[Column, bool]]:
        """
        Casts either the input and output columns to the given inputDtype or
        outputDtype, if specified. Ingress is a boolean that indicates whether we are
        casting the input (True) or output (False) columns.

        :param columns: List of input/output columns to cast.
        :param column_datatypes: List of input/output column datatypes.
        :param ingress: Boolean indicating whether we are casting the input (True) or
        output (False) columns.
        :returns: List of tuple; first element is input/output columns cast to the
        inputDtype/outputDtype, second element is a boolean indicating whether the
        column was cast.
        """
        if ingress:
            cast_dtype = self.getInputDtype()
            if (
                cast_dtype is not None
                and self.compatible_dtypes is not None
                and cast_dtype
                not in [dtype.simpleString() for dtype in self.compatible_dtypes]
            ):
                raise ValueError(
                    f"""inputDtype {cast_dtype} is not a compatible dtype for
                    transformer with uid {self.uid}.
                    Compatible dtypes are{[
                        dtype.simpleString() for dtype in self.compatible_dtypes
                    ]}."""
                )
        else:
            cast_dtype = self.getOutputDtype()

        if cast_dtype is not None:
            return [
                (
                    self._cast(
                        column=column, column_dtype=column_dtype, cast_dtype=cast_dtype
                    ),
                    True,
                )
                if get_element_type(column_dtype).simpleString() != cast_dtype
                else (
                    column,
                    False,
                )
                for column, column_dtype in zip(columns, column_datatypes)
            ]
        return [(column, False) for column in columns]

    def _cast_input_columns(
        self, input_columns: List[Column], input_column_datatypes: List[DataType]
    ) -> List[Tuple[Column, bool]]:
        """
        Casts the input columns to the given inputDtype, if specified. All columns are
        cast to this. This might not be ideal, there may be layers where some inputs are
        expected to be different types. In these cases, the subclass should
        implement the cast_input_columns method.

        :param input_columns: List of input columns to cast.
        :param input_column_datatypes: List of input column datatypes.
        :returns: List of tuple; first element is input columns cast to the inputDtype,
        second element is a boolean indicating whether the column was cast.
        """
        return self._cast_input_output_columns(
            columns=input_columns, column_datatypes=input_column_datatypes, ingress=True
        )

    def _cast_output_columns(
        self, output_columns: List[Column], output_column_datatypes: List[DataType]
    ) -> List[Tuple[Column, bool]]:
        """
        Casts the output columns to the given outputDtype, if specified. All columns are
        cast to this. This might not be ideal, there may be layers where some outputs
        are expected to be different types. In these cases, the subclass should
        implement the cast_output_columns method.

        :param output_columns: List of output columns to cast.
        :param output_column_datatypes: List of output column datatypes.
        :returns: List of tuple; first element is output columns cast to the
        outputDtype, second element is a boolean indicating whether the column was cast.
        """
        return self._cast_input_output_columns(
            columns=output_columns,
            column_datatypes=output_column_datatypes,
            ingress=False,
        )

    def _create_casted_input_output_columns(
        self, dataset: DataFrame, ingress: bool
    ) -> DataFrame:
        """
        Recreates the input or output columns, creating temporary columns with the
        casted input columns if necessary. This is done to avoid overwriting the
        original columns and creating inconsistencies if multiple transforms use the
        same inputs.

        :param dataset: The input dataset.
        :param ingress: Whether the input columns are being cast or the output columns.
        :returns: The dataset with the cast input/output columns.
        """
        col_names = self._get_single_or_multi_col(ingress=ingress)
        columns = [F.col(c) for c in col_names]
        col_datatypes = [dataset.schema[column].dataType for column in col_names]
        casted_cols = (
            self._cast_input_columns(columns, col_datatypes)
            if ingress
            else self._cast_output_columns(columns, col_datatypes)
        )
        for column_name, column_tuple in zip(col_names, casted_cols):
            column = column_tuple[0]
            column_is_cast = column_tuple[1]
            if ingress:
                if column_is_cast:
                    # If we are casting inputs, we create new columns temporarily
                    # to avoid overwriting the original columns and creating
                    # inconsistencies if multiple transforms use the same inputs
                    tmp_input_name = self._resolve_tmp_from_true_column_name(
                        column_name, suffix=self.tmp_column_suffix
                    )
                    dataset = dataset.withColumn(tmp_input_name, column)
            else:
                # Output casting can just be done directly, since this is already a new
                # column created by the transform itself
                if column_is_cast:
                    dataset = dataset.withColumn(column_name, column)
        return dataset

    def drop_tmp_casted_input_columns(self, dataset: DataFrame) -> DataFrame:
        """
        Drops the temporary columns from the dataset that are created when casting the
        input columns.

        :param dataset: The input dataset.
        :returns: The dataset with the temporary columns dropped.
        """
        for column_name in self._get_single_or_multi_col(ingress=True):
            tmp_input_name = self._resolve_tmp_from_true_column_name(
                column_name, suffix=self.tmp_column_suffix
            )
            if tmp_input_name in dataset.columns:
                dataset = dataset.drop(tmp_input_name)
        return dataset

    def set_input_columns_to_from_casted(
        self, dataset: DataFrame, suffix: str, reverse: bool = False
    ) -> None:
        """
        Sets the input columns to the temporary casted columns or back if reverse is
        True.

        :param dataset: The input dataset. Used solely to understand if a tmp column
        has been created. These are only created if the column needed to be cast.
        :param suffix: The suffix to append to the column name.
        :param reverse: Whether to set the input columns back to the original columns.
        :returns: None
        """
        col_names = self._get_single_or_multi_col(ingress=True)
        renamed_columns = [
            self._resolve_tmp_from_true_column_name(column_name, suffix, reverse)
            for column_name in col_names
        ]
        # Check if renaming is necessary. If the renamed column is not present in the
        # dataset, we should not set the input columns to the renamed columns.
        new_col_names = []
        for original_column_name, renamed_column_name in zip(
            col_names, renamed_columns
        ):
            if renamed_column_name in dataset.columns:
                new_col_names.append(renamed_column_name)
            else:
                new_col_names.append(original_column_name)

        if self.hasParam("inputCols") and self.isDefined("inputCols"):
            self._set(inputCols=new_col_names)
        elif self.hasParam("inputCol") and self.isDefined("inputCol"):
            self._set(inputCol=new_col_names[0])
        else:
            raise ValueError("No input columns to set")

    def _check_input_dtypes_compatible(
        self, dataset: DataFrame, column_names: List[str]
    ) -> None:
        """
        Checks if the input tensors are compatible with the compatible_dtypes of the
        layer.

        :param dataset: The input dataset.
        :param column_names: The names of the input columns.
        :raises ValueError: If the input columns are not compatible with the
        compatible_dtypes of the transformer.
        :returns: None
        """
        for c in column_names:
            tmp_column_name = self._resolve_tmp_from_true_column_name(
                c, suffix=self.tmp_column_suffix
            )
            # Either the tmp column name or the original is present.
            check_column_name = (
                tmp_column_name if tmp_column_name in dataset.columns else c
            )
            col_dtype = self.get_column_datatype(dataset, check_column_name)
            underlying_col_dtype = get_element_type(col_dtype)
            if (
                self.compatible_dtypes is not None
                and underlying_col_dtype not in self.compatible_dtypes
            ):
                raise TypeError(
                    f"""Input column with name {c} and dtype
                    {underlying_col_dtype.simpleString()} is not a compatible dtype for
                    transformer with uid {self.uid}.
                    Compatible dtypes are {[
                        dtype.simpleString() for dtype in self.compatible_dtypes
                    ]}."""
                )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import tensorflow as tf
from pyspark.ml import Transformer
from pyspark.sql import DataFrame

from kamae.spark.common import SparkOperation

if TYPE_CHECKING:
    from pyspark.ml._typing import ParamMap


class BaseTransformer(Transformer, SparkOperation):
    """
    Abstract class for all transformers.
    """

    def __init__(self):
        """
        Initializes the transformer.
        """
        super().__init__()

    def transform(
        self, dataset: DataFrame, params: Optional["ParamMap"] = None
    ) -> DataFrame:
        """
        Overrides the transform method of the parent class to add casting of input and
        output columns to the preferred data type.

        :param dataset: Input dataset.
        :param params: Optional additional parameters.
        :returns: Transformed dataset.
        """
        try:
            dataset = self._create_casted_input_output_columns(
                dataset=dataset, ingress=True
            )
            self._check_input_dtypes_compatible(
                dataset, self._get_single_or_multi_col(ingress=True)
            )

            # Set transformer input columns to casted columns
            self.set_input_columns_to_from_casted(
                dataset=dataset,
                suffix=self.tmp_column_suffix,
            )

            # Call the super transform method
            transformed_dataset = super().transform(dataset=dataset, params=params)

            # Reset the transformer input columns from casted columns
            self.set_input_columns_to_from_casted(
                dataset=dataset,
                suffix=self.tmp_column_suffix,
                reverse=True,
            )

            # Drop the temporary casted columns
            transformed_dataset = self.drop_tmp_casted_input_columns(
                transformed_dataset
            )

            transformed_dataset = self._create_casted_input_output_columns(
                dataset=transformed_dataset, ingress=False
            )
            return transformed_dataset
        except Exception as e:
            param_dict = {
                param[0].name: param[1] for param in self.extractParamMap().items()
            }
            raise e.__class__(
                f"Error in transformer: {self.uid} with params: {param_dict}"
            ).with_traceback(e.__traceback__)

    @abstractmethod
    def get_tf_layer(self) -> Union[tf.keras.layers.Layer, List[tf.keras.layers.Layer]]:
        """
        Gets the tensorflow layer to be used in the model.
        This is the only abstract method that must be implemented.
        :returns: Tensorflow Layer
        """
        raise NotImplementedError

    def construct_layer_info(self) -> Dict[str, Any]:
        """
        Constructs the layer info dictionary.
        Contains the layer name, the tensorflow layer, and the inputs and outputs.
        This is used when constructing the pipeline graph.

        :returns: Dictionary containing layer information such as
        name, tensorflow layer, inputs, and outputs.
        """
        inputs, outputs = self.get_layer_inputs_outputs()
        return {
            "name": self.getOrDefault("layerName"),
            "layer": self.get_tf_layer(),
            "inputs": inputs,
            "outputs": outputs,
        }



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List

import tensorflow as tf

from kamae.tensorflow.typing import Tensor


def reshape_to_equal_rank(inputs: Iterable[Tensor]) -> List[Tensor]:
    """
    Reshapes the input tensors to match the rank of the largest tensor.

    :param inputs: The input tensors to reshape.
    :return: The reshaped input tensors.
    """
    max_rank = max([len(tensor.shape) for tensor in inputs])
    reshaped_inputs = []
    for x in inputs:
        rank_diff = max_rank - len(x.shape)
        if rank_diff > 0:
            reshape_dim = tf.concat(
                [
                    tf.shape(x)[:-1],
                    tf.ones(rank_diff, dtype=tf.int32),
                    tf.shape(x)[-1:],
                ],
                axis=0,
            )
            x = tf.reshape(x, reshape_dim)
        reshaped_inputs.append(x)
    return reshaped_inputs



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringToStringListLayer(BaseLayer):
    """
    A layer that converts a string to a list of strings by splitting on a
    separator. It takes a default value and a list_length parameter to ensure that
    the output tensor has the correct shape.

    If the separator is empty, the string is split on bytes/characters.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        separator: str = ",",
        default_value: str = "",
        list_length: int = 1,
        **kwargs,
    ) -> None:
        """
        Initialises the StringToStringListLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param separator: The separator to use when joining the strings.
        Defaults to `","`.
        :param default_value: The value to use when the input is empty.
        Defaults to `""`.
        :param list_length: The length of the string list in the output tensor.
        Defaults to `1`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.separator = separator
        self.list_length = list_length
        self.default_value = default_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Splits the input string tensor by the separator and returns the list of
        strings. A list_length parameter is used to ensure that the output tensor has a
        fixed shape. If the separator is empty, the string is split on bytes/characters.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if an iterable of tensors is passed
        in.

        :param inputs: Input tensor.
        :returns: Tensor with the list of strings.
        """
        input_shape = inputs.get_shape().as_list()
        input_shape.append(self.list_length)
        # If the separator is empty, we split on bytes/characters.
        # Otherwise, we use the standard string split.
        ragged_strings_split = (
            tf.strings.split(inputs, sep=self.separator)
            if self.separator != ""
            else tf.strings.bytes_split(inputs)
        )
        split_strings_tensor = ragged_strings_split.to_tensor(
            default_value=self.default_value, shape=input_shape
        )

        # Replace empty strings with the default value
        split_strings_tensor = tf.where(
            tf.equal(split_strings_tensor, ""), self.default_value, split_strings_tensor
        )

        # If the dimension of the feature was 1, we squeeze it out
        # E.g. (None, None, 1) -> (None, None, 1, N) -> (None, None, N)
        # But (None, None, M) -> (None, None, M, N)
        return (
            tf.squeeze(split_strings_tensor, axis=-2)
            if input_shape[-2] == 1
            else split_strings_tensor
        )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringToStringList layer.
        Used for saving and loading from a model.

        Specifically adds the `axis`, `separator` and `keepdims` to the config
        dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "separator": self.separator,
                "default_value": self.default_value,
                "list_length": self.list_length,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import numpy as np
import pandas as pd
import tensorflow as tf

from kamae.sklearn.params import MultiInputSingleOutputMixin
from kamae.tensorflow.layers import ArrayConcatenateLayer

from .base import BaseTransformer


class ArrayConcatenateTransformer(
    BaseTransformer,
    MultiInputSingleOutputMixin,
):
    """
    Vector Assembler Scikit-Learn Transformer for use in Scikit-Learn pipelines.
    This transformer assembles multiple columns into a single array column.
    """

    def __init__(self, input_cols: List[str], output_col: str, layer_name: str) -> None:
        super().__init__()
        self.input_cols = input_cols
        self.output_col = output_col
        self.layer_name = layer_name

    def fit(self, X: pd.DataFrame, y=None) -> "ArrayConcatenateTransformer":
        """
        Fits the transformer to the data. Does nothing since
        this is transformer not an estimator.

        :param X: Pandas dataframe to fit the transformer to.
        :param y: Not used, present here for API consistency by convention.
        :returns: Fit pipeline, in this case the transformer itself.
        """
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform the input dataset. Creates a new column named outputCol which is a
        concatenated array of all input columns.

        :param X: Pandas dataframe to transform.
        :param y: Not used, present here for API consistency by convention.
        :returns: Transformed data.
        """

        # Check which columns are arrays, this gives a dict like:
        # {'col1': True, 'col2': False, 'col3': True}
        is_col_an_array_dict = (
            X.head(1)[self.input_cols]
            .applymap(lambda x: pd.api.types.is_list_like(x))
            .to_dict(orient="records")[0]
        )

        new_input_cols = []
        for col_name, col_an_array in is_col_an_array_dict.items():
            if col_an_array:
                # If the column is an array then we need to create a
                # numpy array of arrays
                # TODO: Can we make this more this efficient?
                values = X[col_name].to_numpy()
                new_input_cols.append(np.array([np.array(x) for x in values]))
            else:
                # If the column is not an array then we just need to extend
                # the numpy array to have an extra dimension. This is so we can concat
                # the arrays later.
                values = X[col_name].to_numpy()
                new_input_cols.append(values[:, None])

        # Concatenate the arrays, this creates an N x M array
        # where N is the number of rows, M is the number of features
        concatenated_array = np.concatenate(new_input_cols, axis=-1)
        # Add this to the dataframe, convert the numpy array to a list
        # of 1-D numpy arrays
        X[self.output_col] = list(concatenated_array)

        return X

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that concatenates the input tensors.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that concatenates the input tensors.
        """
        return ArrayConcatenateLayer(name=self.layer_name, axis=-1)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional, Union

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import ImputeLayer

from .base import BaseTransformer


class ImputeParams(Params):
    """
    Mixin class used to provide imputation and mask value needed for imputation.
    """

    imputeValue = Param(
        Params._dummy(),
        "imputeValue",
        "Value to be imputed.",
        typeConverter=TypeConverters.identity,
    )

    maskValue = Param(
        Params._dummy(),
        "maskValue",
        """
        Value which is to be replaced with the imputation value.
        Value is ignored when computing the imputation value.
        """,
        typeConverter=TypeConverters.identity,
    )

    def setImputeValue(self, value: Union[float, int, str]) -> "ImputeParams":
        """
        Sets the parameter imputeValue to the given scalar value.

        :param value: List of imputeValues values.
        :returns: Instance of class mixed in.
        """
        if value is None:
            raise ValueError("Impute value cannot be None")
        return self._set(imputeValue=value)

    def getImputeValue(self) -> Union[float, int, str]:
        """
        Gets the imputeValue parameter.

        :returns: String, float or int imputeValue
        """
        return self.getOrDefault(self.imputeValue)

    def setMaskValue(self, value: Union[float, int, str]) -> "ImputeParams":
        """
        Sets the maskValue parameter.

        :param value: Float value to use as the mask value.
        :returns: Instance of class mixed in.
        """
        return self._set(maskValue=value)

    def getMaskValue(self) -> Union[float, int, str]:
        """
        Gets the maskValue parameter.
        :returns: String, float or int value of the mask value.
        """
        return self.getOrDefault(self.maskValue)


class ImputeTransformer(BaseTransformer, ImputeParams, SingleInputSingleOutputParams):
    """
    Imputation transformer for use in Spark pipelines.
    This is used to impute the mean or median value when
    value is null or equalling a mask
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        imputeValue: Optional[Union[float, int, str]] = None,
        maskValue: Optional[Union[float, int, str]] = None,
    ) -> None:
        """
        Initializes an ImputeTransformer transformer.

        :param inputCol: Input column name to impute over.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model.
        :param imputeValue: String, float or int value to impute in place of mask or
        nulls.
        :param maskValue: String, float or int value which should be ignored when
        computing the imputation statistic and to be replaced by the imputation.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input data by imputing the specified
        statistic of the input column provided.
        Nulls and any values equalling the mask are imputed over.

        :param dataset: Pyspark dataframe to transform.
        :returns: Pyspark dataframe with the input columns
        imputed with the specified statistic, and names set to
        output column names provided
        """
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: F.when(
                (x == F.lit(self.getMaskValue())) | (x.isNull()),
                F.lit(self.getImputeValue()),
            ).otherwise(x),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the imputation transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
         that performs the imputation.
        """
        mask_value = self.getMaskValue()

        return ImputeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            impute_value=self.getImputeValue(),
            mask_value=mask_value,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(kamae.__name__)
class ArrayCropLayer(BaseLayer):
    """
    Performs a cropping of the input tensor to a certain length.
    If the tensor is shorter than the specified length, it is
    padded with specified pad value.

    TODO: Currently only supports cropping the final dimension of the tensor.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: Union[str, int, float] = None,
        output_dtype: Union[str, int, float] = None,
        array_length: int = 128,
        pad_value: Union[str, int, float] = None,
        **kwargs,
    ) -> None:
        """
        Initialises the ArrayCropLayer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param array_length: The length to crop or pad the arrays to. Defaults to 128.
        :param pad_value: The value to pad the arrays with. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if array_length < 1:
            raise ValueError("Array length must be greater than 0.")
        self.array_length = array_length

        if pad_value is None:
            raise ValueError("Pad value must be provided and not None.")
        self.pad_value = pad_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Crops the tensor to specified length and pads with specified value.

        :param inputs: Tensor to split.
        :returns: Cropped and padded tensor
        """
        inputs_shape = tf.shape(inputs)

        # Crop final dimension of tensor
        crop_length = tf.minimum(self.array_length, inputs_shape[-1])
        cropped = inputs[..., :crop_length]

        # Pad final dim of tensor if necessary
        padding_length = tf.maximum(self.array_length - inputs_shape[-1], 0)
        paddings = [[0, 0]] * (inputs_shape.shape[0] - 1) + [[0, padding_length]]
        padded = tf.pad(cropped, paddings, constant_values=self.pad_value)
        new_shape = tf.concat(
            [
                tf.shape(padded)[:-1],
                tf.expand_dims(tf.constant(self.array_length), axis=-1),
            ],
            axis=0,
        )
        return tf.reshape(padded, new_shape)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the ArrayCrop layer.
        Used for saving and loading from a model.

        Specifically, adds the `array_length` amd `pad_value to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"array_length": self.array_length, "pad_value": self.pad_value})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import ArrayType, DataType, IntegerType, StringType

from kamae.spark.params import HashIndexParams, SingleInputSingleOutputParams
from kamae.spark.utils import (
    hash_udf,
    single_input_single_output_array_udf_transform,
    single_input_single_output_scalar_transform,
)
from kamae.tensorflow.layers import BloomEncodeLayer

from .base import BaseTransformer


class BloomEncodeParams(HashIndexParams):
    """
    Mixin class containing parameters needed for bloom encoding.
    """

    numHashFns = Param(
        Params._dummy(),
        "numHashFns",
        "Number of hash functions to use for bloom encoding",
        typeConverter=TypeConverters.toInt,
    )

    featureCardinality = Param(
        Params._dummy(),
        "featureCardinality",
        "Dimension/cardinality of the feature",
        typeConverter=TypeConverters.toInt,
    )

    useHeuristicNumBins = Param(
        Params._dummy(),
        "useHeuristicNumBins",
        "Whether to use te heuristic from the paper to determine the number of bins",
        typeConverter=TypeConverters.toBoolean,
    )

    def setNumHashFns(self, value: int) -> "BloomEncodeParams":
        """
        Sets the `numHashFns` parameter.
        """
        if value < 2:
            raise ValueError("numHashFns must be at least 2.")
        return self._set(numHashFns=value)

    def getNumHashFns(self) -> int:
        """
        Gets the value of `numHashFns` parameter.
        """
        return self.getOrDefault(self.numHashFns)

    def setFeatureCardinality(self, value: int) -> "BloomEncodeParams":
        """
        Sets the `featureCardinality` parameter.
        """
        if value < 1:
            raise ValueError("featureCardinality must be greater than 0")
        return self._set(featureCardinality=value)

    def getFeatureCardinality(self) -> int:
        """
        Gets the value of `featureCardinality` parameter.
        """
        return self.getOrDefault(self.featureCardinality)

    def setUseHeuristicNumBins(self, value: bool) -> "BloomEncodeParams":
        """
        Sets the `useHeuristicNumBins` parameter.
        """
        return self._set(useHeuristicNumBins=value)

    def getUseHeuristicNumBins(self) -> bool:
        """
        Gets the value of `useHeuristicNumBins` parameter.
        """
        return self.getOrDefault(self.useHeuristicNumBins)

    def getNumBins(self) -> int:
        """
        Gets the number of bins to use for hash indexing.
        """
        if self.getUseHeuristicNumBins() and self.getFeatureCardinality() is not None:
            return max(round(self.getFeatureCardinality() * 0.2), 2)
        elif self.getUseHeuristicNumBins():
            raise ValueError(
                """If useHeuristicNumBins is set to True, then the featureCardinality
                parameter must be set."""
            )
        return self.getOrDefault(self.numBins)


class BloomEncodeTransformer(
    BaseTransformer,
    BloomEncodeParams,
    SingleInputSingleOutputParams,
):
    """
    Bloom encoder Spark Transformer for use in Spark pipelines.
    This transformer performs bloom encoding on the input column resulting in an
    array of integers of size equal to numHashFns.
    See paper for more details: https://arxiv.org/pdf/1706.03993.pdf
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        numHashFns: int = 3,
        numBins: Optional[int] = None,
        maskValue: Optional[str] = None,
        featureCardinality: Optional[int] = None,
        useHeuristicNumBins: bool = False,
    ) -> None:
        """
        Instantiates a BloomEncode transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param numHashFns: Number of hash functions to use. Defaults to 3.
        The paper suggests a range of 2-4 hash functions for optimal performance.
        :param numBins: Number of hash bins. Note that this includes the `maskValue`
        bin, so the effective number of bins is `(num_bins - 1)` if `maskValue`
        is set. If `useHeuristicNumBins` is set to True, then this parameter is
        ignored and the number of bins is automatically set. See the description of this
        parameter below for how the heuristic is built.
        :param maskValue: A value that represents masked inputs, which are mapped to
        index 0. Defaults to None, meaning no mask term will be added and the
        hashing will start at index 0.
        :param featureCardinality: The cardinality of the input tensor. Needed to
        use the num of bins heuristic. Defaults to None, meaning the number of bins will
        not use the heuristic and will need to be set manually.
        :param useHeuristicNumBins: If set to True, the number of bins is automatically
        set by fixing the ratio of the feature cardinality to the number of bins
        to be b/f = 0.2. This ratio was found to be optimal in the paper for a wide
        variety of usecases. Therefore, numBins = featureCardinality * 0.2. This reduces
        the cardinality of the input tensor by 5x.
        Requires the `featureCardinality` parameter to be set. Defaults to False.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self._setDefault(
            numHashFns=3,
            numBins=None,
            maskValue=None,
            featureCardinality=None,
            useHeuristicNumBins=False,
        )
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _create_salted_input(self, column_data_type: DataType) -> Column:
        """
        Builds the salted inputs according to how many hash functions are used.
        Specifically concatenates the input column with the string "0" using a
        separator of the hash function index.

        :returns: Salted input spark column
        """
        return single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=column_data_type,
            func=lambda x: F.array(
                [F.concat(F.lit(x), F.lit(i)) for i in range(self.getNumHashFns())]
            ),
        )

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column named outputCol with the
        bloom encoded input column.

        :param dataset: Pyspark DataFrame to transform.
        :returns: Transformed pyspark dataFrame.
        """
        num_bins = self.getNumBins()
        if num_bins is None:
            # num_bins can be None only if useHeuristicNumBins is False
            raise ValueError("numBins must be set if useHeuristicNumBins is False.")
        mask_value = self.getMaskValue()

        input_data_type = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        salted_input = self._create_salted_input(input_data_type)

        # The salting process nests the input column into another array. Thus, the array
        # nesting level is increased by 1.
        def bloom_encode(x):
            return [
                hash_udf(
                    label=y,
                    num_bins=num_bins,
                    # If the user set a mask value, then this won't match the inputs
                    # as they have been salted. So we need to salt the mask value as
                    # well.
                    mask_value=f"{mask_value}{i}" if mask_value is not None else None,
                )
                for i, y in enumerate(x)
            ]

        output_col = single_input_single_output_array_udf_transform(
            input_col=salted_input,
            # Input datatype is from salted input, so it has an additional; nesting.
            input_col_datatype=ArrayType(input_data_type),
            func=bloom_encode,
            udf_return_element_datatype=IntegerType(),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the bloom encoding.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the bloom encoding operation.
        """
        return BloomEncodeLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            num_hash_fns=self.getNumHashFns(),
            num_bins=self.getNumBins(),
            mask_value=self.getMaskValue(),
            feature_cardinality=self.getFeatureCardinality(),
            use_heuristic_num_bins=self.getUseHeuristicNumBins(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

from kamae.spark.utils.indexer_utils import safe_hash64


def hash_udf(label: str, num_bins: int, mask_value: str = None) -> Union[int, None]:
    """
    User defined Spark function (UDF) to hash a string to an integer value.

    This uses FarmHash64. We have a python binding of the Google library
    https://github.com/google/farmhash within the pyfarmhash package.

    Therefore, this matches the hash function used in the tensorflow layer.

    :param label: String to hash.
    :param num_bins: Number of bins to hash to.
    :param mask_value: Mask value to use for hash indexing.
    :returns: Hashed integer value.
    """
    if label is None:
        return None
    if label == mask_value:
        return 0

    hash_val = safe_hash64(label)
    if mask_value is not None:
        # If masking value is set, then the zero index is reserved for it.
        # Therefore, we reduce the num_bins by 1 and add 1 to the binned value.
        return (hash_val % (num_bins - 1)) + 1
    else:
        return hash_val % num_bins


def indexer_udf(
    label: str,
    labels: List[str],
    mask_token: Optional[str] = None,
    num_oov_indices: int = 1,
) -> int:
    """
    User defined Spark function (UDF) to index a label based on the labels array
    mask token, and number of out of vocabulary indices.

    If the label is the mask token, we return 0. If the label is not in the labels
    array and num_oov_indices is 0, we raise an error. Otherwise, we calculate the
    offset for the index based on the mask token and num_oov_indices. Lastly, if
    the number of out of vocabulary indices is more than 1 we hash the label to an
    out of vocabulary index.

    Raises an error if the provided label is hashed and contains a null character.

    :param label: Current label to index.
    :param labels: Array of labels to index.
    :param mask_token: Mask token to use for the 0 index.
    :param num_oov_indices: Number of out of vocabulary indices.
    :returns: Indexed integer value.
    """
    if label is None:
        if num_oov_indices > 0:
            return 0
        else:
            raise ValueError(
                """Found null label but numOOVIndices is 0.
                Consider setting numOOVIndices to 1 or more to cater for null labels."""
            )

    if mask_token is not None and label == mask_token:
        # If the label is the mask token,
        # we want to return 0, since 0 is reserved for masked tokens
        return 0

    # Calculate the offset for the index
    start_index = 1 if mask_token is not None else 0
    offset = num_oov_indices + start_index

    # Attempt to get the index.
    try:
        return labels.index(label) + offset
    except ValueError:
        if num_oov_indices == 0:
            # If we have no out of vocabulary indices and the label is not in the
            # labels array, raise an error.
            raise ValueError(
                f"""Label {label} not found in labels array and
                numOOVIndices is 0. Consider setting numOOVIndices to 1 or more."""
            )
        # If the label is not in the labels array and not equal to the mask token,
        # then we need to hash it to an out of vocabulary index.
        hashed_int = safe_hash64(label)
        return (hashed_int % num_oov_indices) + start_index


def one_hot_encoding_udf(
    label: str,
    labels: List[str],
    mask_token: Optional[str] = None,
    num_oov_indices: int = 1,
    drop_unseen: bool = False,
) -> List[float]:
    """
    User defined Spark function (UDF) to create a one-hot encoding of a label based
    on the labels array, mask token, number of out of vocabulary indices, and whether
    to drop unseen labels.

    :param label: Current label to index.
    :param labels: Array of labels to index.
    :param mask_token: Mask token to use for the 0 index.
    :param num_oov_indices: Number of out of vocabulary indices.
    :param drop_unseen: Whether to drop unseen label indices.
    :returns: List of floats representing the one-hot encoding.
    """
    index = indexer_udf(
        label=label,
        labels=labels,
        mask_token=mask_token,
        num_oov_indices=num_oov_indices,
    )
    mask_offset = 1 if mask_token is not None else 0
    if drop_unseen:
        encoding = [0.0] * len(labels)
        shifted_index = index - (num_oov_indices + mask_offset)
        if shifted_index >= 0:
            encoding[shifted_index] = 1.0
    else:
        encoding = [0.0] * (len(labels) + num_oov_indices + mask_offset)
        encoding[index] = 1.0
    return encoding


def ordinal_array_encode_udf(
    value: List[str], pad_value: Optional[str] = None
) -> List[int]:
    """
    User defined Spark function (UDF) to encode a list of strings as an ordinal array.
    Example:
    value = ['a', 'b', 'c']
    ordinal_array = [0, 1, 2]

    :param value: List of strings to encode.
    :param pad_value: Value to use for padding. Padded values get an index of -1.
    :returns: List of integers representing the ordinal array.
    """
    string_index_mapping = {pad_value: -1}
    ordinal_array = []
    for string in value:
        if string not in string_index_mapping:
            string_index_mapping[string] = len(string_index_mapping) - 1
        ordinal_array.append(string_index_mapping[string])
    return ordinal_array



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import (
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
)

from kamae.spark.params import (
    MathFloatConstantParams,
    MultiInputSingleOutputParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import MinLayer

from .base import BaseTransformer


class MinTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    MathFloatConstantParams,
):
    """
    MinLayer Spark Transformer for use in Spark pipelines.
    This transformer gets the min of a column and a constant or another column.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        mathFloatConstant: Optional[float] = None,
    ) -> None:
        """
        Initializes an MinTransformer transformer.

        :param inputCol: Input column name. Only used if inputCols is not specified.
        If specified, we min this column by the mathFloatConstant.
        :param inputCols: Input column names.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param mathFloatConstant: Optional constant to use for min op. If not provided,
        then two input columns are required.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(mathFloatConstant=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
            ByteType(),
            ShortType(),
            IntegerType(),
            LongType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the minimum of either the `inputCols` if specified, or the `inputCol`
        and the `mathFloatConstant`

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self.get_multiple_input_cols(
            constant_param_name="mathFloatConstant"
        )
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: F.least(*[x[c] for c in input_col_names]),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the min transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a min operation.
        """
        return MinLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            min_constant=self.getMathFloatConstant(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import tensorflow as tf

from kamae.sklearn.params import SingleInputSingleOutputMixin
from kamae.tensorflow.layers import IdentityLayer

from .base import BaseTransformer


class IdentityTransformer(BaseTransformer, SingleInputSingleOutputMixin):
    """
    Identity Scikit-Learn Transformer for use in Scikit-Learn pipelines.
    This transformer simply passes the input to the output unchanged.
    Used for cases where you want to keep the input the same.
    """

    def __init__(self, input_col: str, output_col: str, layer_name: str) -> None:
        """
        Intializes an IdentityTransformer transformer.

        :param input_col: Input column name.
        :param output_col: Output column name.
        :param layer_name: Name of the layer. Used as the name of the tensorflow layer
        in the keras model.
        :returns: None - class instantialized.
        """
        super().__init__()
        self.input_col = input_col
        self.output_col = output_col
        self.layer_name = layer_name

    def fit(self, X: pd.DataFrame, y=None) -> "IdentityTransformer":
        """
        Fits the transformer to the data. Does nothing since
        this is an identity transformer.

        :param X: Pandas dataframe to fit the transformer to.
        :param y: Not used, present here for API consistency by convention.
        :returns: Fit pipeline, in this case the transformer itself.
        """
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transforms the data using the transformer. Creates a new column with name
        `output_col`, which is the same as the `input_col`.

        :param X: Pandas dataframe to transform.
        :param y: Not used, present here for API consistency by convention.
        :returns: Transformed data.
        """
        X[self.output_col] = X[self.input_col]
        return X

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the identity transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs an Identity operation.
        """
        return IdentityLayer(
            name=self.layer_name,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class RoundToDecimalLayer(BaseLayer):
    """
    Performs a rounding to the nearest decimal operation on the input tensor.

    If the specified number of decimals is too large for the input precision type,
    this operation can result in overflow. This is because the operation is performed by
    multiplying the input tensor by 10 to the power of the number of decimals, rounding
    the result to the nearest integer, and then dividing by 10 to the power of the
    number of decimals.
    """

    def __init__(
        self,
        decimals: int = 1,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the RoundToDecimalLayer layer.

        :param decimals: The number of decimal places to round to.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if decimals < 0:
            raise ValueError("""decimals must be greater than or equal to 0.""")
        self.decimals = decimals

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.float16, tf.float32, tf.float64, tf.int32, tf.int64]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the rounding operation on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input is a
        single tensor. Raises an error if multiple tensors are passed in as an iterable.

        :param inputs: Input tensor to perform the rounding on.
        :returns: The input tensor with the rounding applied.
        """
        # WARNING: Depending on the type of the input and the number of decimals,
        # this multiplier could overflow.
        max_val = inputs.dtype.max
        if 10**self.decimals > max_val:
            raise ValueError(
                """The number of decimals is too large for the input dtype.
                Overflow expected."""
            )
        multiplier = tf.constant(10**self.decimals, dtype=inputs.dtype)
        return tf.round(inputs * multiplier) / multiplier

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the RoundToDecimal layer.
        Used for saving and loading from a model.

        Specifically adds the `decimals` value to the configuration.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"decimals": self.decimals})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List, Union

import numpy as np
import tensorflow as tf

from kamae.tensorflow.typing import Tensor


def get_top_n(
    val_tensor: Tensor,
    axis: int,
    sort_tensor: Tensor,
    top_n: int,
    sort_order: str = "asc",
) -> Tensor:
    """
    Get the top N items from the value tensor based on their position in
    the sort tensor, ordered by the sort order ('asc' or 'desc').

    :param val_tensor: Value tensor.
    :param axis: Axis to get the top N items.
    :param sort_tensor: Sort tensor.
    :param top_n: Number of top values to consider.
    :param sort_order: Order to sort the values by. Default is "asc".
    :returns: Tensor of the top N items
    """

    # If K is less than the number of items at real time,
    # replace K with the number of items in the list
    top_n = tf.minimum(top_n, tf.shape(sort_tensor)[axis])

    # Define sort direction
    sort_tensor_with_order = None
    if sort_order == "desc":
        sort_tensor_with_order = sort_tensor
    elif sort_order == "asc":
        sort_tensor_with_order = -sort_tensor
    else:
        ValueError(f"Invalid sort_order: {sort_order}")

    # If value of shape at position (axis + 1) is equal to 1, squeeze this dimension,
    # otherwise the top_k would complain about the shape mismatch
    # If we apply squeeze without axis, the inference when batch_size=1 would fail
    if len(sort_tensor_with_order.shape) > axis + 1:
        if sort_tensor_with_order.shape[axis + 1] == 1:
            sort_tensor_with_order = tf.squeeze(sort_tensor_with_order, axis=axis + 1)

    # Get the indices of the top N items, using the sort tensor
    _, sorted_indices = tf.math.top_k(sort_tensor_with_order, k=top_n, sorted=True)

    # Gather elements from the value tensor using the top-k indices
    return tf.gather(
        val_tensor,
        sorted_indices,
        batch_dims=axis,
        axis=axis,
    )


def listify_tensors(x: Union[tf.Tensor, np.ndarray, List[Any]]) -> List[Any]:
    """
    Converts any tensors or numpy arrays to lists for config serialization.

    :param x: The input tensor or numpy array.
    :returns: The input as a list.
    """
    if tf.is_tensor(x):
        x = x.numpy()
    if isinstance(x, np.ndarray):
        x = x.tolist()
    return x



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
import math
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType

from kamae.spark.params import LatLonConstantParams, MultiInputSingleOutputParams
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import BearingAngleLayer

from .base import BaseTransformer


class BearingAngleParams(LatLonConstantParams, MultiInputSingleOutputParams):
    """
    Mixin class setting input cols.
    """

    def setInputCols(self, value: List[str]) -> "BearingAngleParams":
        """
        Overrides setting the input columns for the transformer.
        Throws an error if we do not have either two or four input columns depending
        on whether latLonConstant is provided.
        :param value: List of input columns.
        :returns: Class instance.
        """
        if self.getLatLonConstant() is not None and len(value) != 2:
            raise ValueError(
                """When setting inputCols for HaversineDistanceTransformer,
                if the latLonConstant is not None,
                there must be exactly two input columns."""
            )
        elif len(value) not in [2, 4]:
            raise ValueError(
                """When setting inputCols for HaversineDistanceTransformer,
                there must be either two or four input columns."""
            )
        return self._set(inputCols=value)


class BearingAngleTransformer(
    BaseTransformer,
    BearingAngleParams,
):
    """
    Bearing Angle Calculator Spark Transformer for use in Spark pipelines.
    This transformer computes the bearing angle between two lat/lon pairs.
    This can be between four columns (one for each lat/lon) or between two columns
    and a constant.
    The transformer will return null angle if any of the lat/lon values
    are out of bounds. For lat, this is [-90, 90] and for lon, this is [-180, 180].
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        latLonConstant: Optional[List[float]] = None,
    ) -> None:
        """
        Initializes a BearingAngle transformer.
        :param inputCols: Input column names. If latLonConstant is provided, then two
        input columns are required. These must be in the order [lat, lon].
        If latLonConstant is not provided, then four input columns are required.
        These must be in the order [lat1, lon1, lat2, lon2].
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param latLonConstant: Optional list of lat/lon constant to use.
        Must be in the order [lat, lon].
        If not provided, then four input columns are required.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(latLonConstant=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.
        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    @staticmethod
    def validate_lat_lon_column(x: Column, lat: bool = True) -> Column:
        """
        Validates that the input column is a valid lat or lon column.
        If not, then casts the column to null.
        :param x: Input column.
        :param lat: Whether the column is a lat column or not.
        :returns: Column.
        """
        cond = x.between(-90.0, 90.0) if lat else x.between(-180.0, 180.0)
        return F.when(cond, x)

    def _get_input_cols(self) -> List[Column]:
        """
        Gets the input columns as a list of pyspark.sql.Column. Also checks if the
        lat and long are out of bounds and if so casts them to null.
        This will make the output null.
        If the latLonConstant is provided, then returns the input columns and the
        constant split into lat and lon. Otherwise, returns the input columns.
        :returns: List of input columns.
        """
        input_cols = self.getInputCols()

        lat_lon_constant = self.getLatLonConstant()

        if lat_lon_constant is not None:
            return [
                F.col(input_cols[0]).alias(self.uid + "_lat1"),
                F.col(input_cols[1]).alias(self.uid + "_lon1"),
                F.lit(lat_lon_constant[0]).alias(self.uid + "_lat2"),
                F.lit(lat_lon_constant[1]).alias(self.uid + "_lon2"),
            ]
        else:
            return [
                F.col(input_cols[0]).alias(self.uid + "_lat1"),
                F.col(input_cols[1]).alias(self.uid + "_lon1"),
                F.col(input_cols[2]).alias(self.uid + "_lat2"),
                F.col(input_cols[3]).alias(self.uid + "_lon2"),
            ]

    @staticmethod
    def _to_radians_col(x: Column) -> Column:
        """
        Converts a column of degrees to radians.
        :param x: Column of degrees.
        :returns: Column of radians in double precision.
        """
        return x.cast(DoubleType()) * F.lit(math.pi / 180)

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the bearing angle between the input lat/lon columns.
        Returns null if any of the lat/lon values are out of bounds.
        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self._get_input_cols()
        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]

        def bearing_calculate_transform(
            x: Column, input_col_names: List[str]
        ) -> Column:
            lat1_radians = self._to_radians_col(
                self.validate_lat_lon_column(x[input_col_names[0]])
            )
            lon1_radians = self._to_radians_col(
                self.validate_lat_lon_column(x[input_col_names[1]], lat=False)
            )
            lat2_radians = self._to_radians_col(
                self.validate_lat_lon_column(x[input_col_names[2]])
            )
            lon2_radians = self._to_radians_col(
                self.validate_lat_lon_column(x[input_col_names[3]], lat=False)
            )

            lon_difference = lon2_radians - lon1_radians
            # Bearing formula calculation
            y = F.sin(lon_difference) * F.cos(lat2_radians)

            x = F.cos(lat1_radians) * F.sin(lat2_radians)
            x -= F.sin(lat1_radians) * F.cos(lat2_radians) * F.cos(lon_difference)

            # Calculate bearing in degrees
            bearing = F.atan2(y, x)
            bearing_deg = F.pmod(F.degrees(bearing) + 360, 360)
            return bearing_deg

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: bearing_calculate_transform(x, input_col_names),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the bearing angle transformer.
        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         computes the bearing angle between two lat/lon pairs.
        """
        return BearingAngleLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            lat_lon_constant=self.getLatLonConstant(),
        )
